@@ -25,11 +25,10 @@ import (
 	"time"
 
 	bmh "github.com/metal3-io/baremetal-operator/pkg/apis/metal3/v1alpha1"
-	bmv1alpha1 "github.com/metal3-io/cluster-api-provider-baremetal/pkg/apis/baremetal/v1alpha1"
-	clusterv1 "github.com/openshift/cluster-api/pkg/apis/cluster/v1alpha1"
-	"github.com/openshift/cluster-api/pkg/apis/machine/common"
-	machinev1 "github.com/openshift/cluster-api/pkg/apis/machine/v1beta1"
-	clustererror "github.com/openshift/cluster-api/pkg/controller/error"
+	"github.com/metal3-io/baremetal-operator/pkg/utils"
+	bmv1alpha1 "github.com/openshift/cluster-api-provider-baremetal/pkg/apis/baremetal/v1alpha1"
+	machinev1beta1 "github.com/openshift/machine-api-operator/pkg/apis/machine/v1beta1"
+	machineapierrors "github.com/openshift/machine-api-operator/pkg/controller/machine"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/equality"
 	"k8s.io/apimachinery/pkg/api/errors"
@@ -45,8 +44,11 @@ const (
 	ProviderName = "baremetal"
 	// HostAnnotation is the key for an annotation that should go on a Machine to
 	// reference what BareMetalHost it corresponds to.
-	HostAnnotation = "metal3.io/BareMetalHost"
-	requeueAfter   = time.Second * 30
+	HostAnnotation                = "metal3.io/BareMetalHost"
+	requeueAfter                  = time.Second * 30
+	externalRemediationAnnotation = "host.metal3.io/external-remediation"
+	poweredOffForRemediation      = "remediation.metal3.io/powered-off-for-remediation"
+	requestPowerOffAnnotation     = "reboot.metal3.io/capbm-requested-power-off"
 )
 
 // Add RBAC rules to access cluster-api resources
@@ -75,7 +77,7 @@ func NewActuator(params ActuatorParams) (*Actuator, error) {
 }
 
 // Create creates a machine and is invoked by the Machine Controller
-func (a *Actuator) Create(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
+func (a *Actuator) Create(ctx context.Context, machine *machinev1beta1.Machine) error {
 	log.Printf("Creating machine %v .", machine.Name)
 
 	// load and validate the config
@@ -106,25 +108,31 @@ func (a *Actuator) Create(ctx context.Context, cluster *clusterv1.Cluster, machi
 
 	// none found, so try to choose one
 	if host == nil {
-		host, err = a.chooseHost(ctx, machine, config)
+		host, err = a.chooseHost(ctx, machine)
 		if err != nil {
 			return err
 		}
 		if host == nil {
 			log.Printf("No available host found. Requeuing.")
-			return &clustererror.RequeueAfterError{RequeueAfter: requeueAfter}
+			return &machineapierrors.RequeueAfterError{RequeueAfter: requeueAfter}
 		}
 		log.Printf("Associating machine %s with host %s", machine.Name, host.Name)
 	} else {
 		log.Printf("Machine %s already associated with host %s", machine.Name, host.Name)
 	}
 
+	// Add a finalizer for the BMH. This will allow us to delete
+	// the Machine when the BMH is deleted.
+	if !utils.StringInList(host.Finalizers, machinev1beta1.MachineFinalizer) {
+		host.Finalizers = append(host.Finalizers, machinev1beta1.MachineFinalizer)
+	}
+
 	err = a.setHostSpec(ctx, host, machine, config)
 	if err != nil {
 		return err
 	}
 
-	err = a.ensureAnnotation(ctx, machine, host)
+	_, err = a.ensureAnnotation(ctx, machine, host)
 	if err != nil {
 		return err
 	}
@@ -138,7 +146,7 @@ func (a *Actuator) Create(ctx context.Context, cluster *clusterv1.Cluster, machi
 }
 
 // Delete deletes a machine and is invoked by the Machine Controller
-func (a *Actuator) Delete(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
+func (a *Actuator) Delete(ctx context.Context, machine *machinev1beta1.Machine) error {
 	log.Printf("Deleting machine %v .", machine.Name)
 	host, err := a.getHost(ctx, machine)
 	if err != nil {
@@ -159,7 +167,7 @@ func (a *Actuator) Delete(ctx context.Context, cluster *clusterv1.Cluster, machi
 			if err != nil && !errors.IsNotFound(err) {
 				return err
 			}
-			return &clustererror.RequeueAfterError{}
+			return &machineapierrors.RequeueAfterError{}
 		}
 
 		waiting := true
@@ -175,7 +183,7 @@ func (a *Actuator) Delete(ctx context.Context, cluster *clusterv1.Cluster, machi
 			waiting = host.Status.PoweredOn
 		}
 		if waiting {
-			return &clustererror.RequeueAfterError{RequeueAfter: requeueAfter}
+			return &machineapierrors.RequeueAfterError{RequeueAfter: requeueAfter}
 		} else {
 			host.Spec.ConsumerRef = nil
 			err = a.client.Update(ctx, host)
@@ -189,7 +197,7 @@ func (a *Actuator) Delete(ctx context.Context, cluster *clusterv1.Cluster, machi
 }
 
 // Update updates a machine and is invoked by the Machine Controller
-func (a *Actuator) Update(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) error {
+func (a *Actuator) Update(ctx context.Context, machine *machinev1beta1.Machine) error {
 	log.Printf("Updating machine %v .", machine.Name)
 
 	// clear any error message that was previously set. This method doesn't set
@@ -204,14 +212,53 @@ func (a *Actuator) Update(ctx context.Context, cluster *clusterv1.Cluster, machi
 		return err
 	}
 	if host == nil {
+		// When finalizer for BHM is removed but an error occur before
+		// the Machine is deleted below in the StateDeleting block, the
+		// Machine should be deleted when no host is found.
+		log.Print("Deleting machine whose associated host is gone: ", machine.Name)
+		err = a.client.Delete(ctx, machine)
+		if err != nil && !errors.IsNotFound(err) {
+			return err
+		}
+		log.Print("Deleted machine whose associated host is gone: ", machine.Name)
 		return fmt.Errorf("host not found for machine %s", machine.Name)
 	}
 
-	err = a.ensureAnnotation(ctx, machine, host)
+	// Delete Machine when BareMetalHost is deleted.
+	// This is to ensure the MachineSet creates a new Machine
+	// containing the latest ProviderSpec.
+	if host.Status.Provisioning.State == bmh.StateDeleting {
+		if utils.StringInList(host.Finalizers, machinev1beta1.MachineFinalizer) {
+			log.Print("Removing finalizer for host: ", host.Name)
+			host.Finalizers = utils.FilterStringFromList(
+				host.Finalizers, machinev1beta1.MachineFinalizer)
+			err = a.client.Update(ctx, host)
+			if err != nil && !errors.IsNotFound(err) {
+				return err
+			}
+			log.Print("Removed finalizer for host: ", host.Name)
+		}
+
+		log.Print("Deleting machine whose associated host is gone: ", machine.Name)
+		err = a.client.Delete(ctx, machine)
+		if err != nil && !errors.IsNotFound(err) {
+			return err
+		}
+		log.Print("Deleted machine whose associated host is gone: ", machine.Name)
+		return nil
+	}
+
+	dirty, err := a.ensureAnnotation(ctx, machine, host)
 	if err != nil {
 		return err
 	}
 
+	if !dirty {
+		if err := a.remediateIfNeeded(ctx, machine, host); err != nil {
+			return err
+		}
+	}
+
 	if err := a.updateMachineStatus(ctx, machine, host); err != nil {
 		return err
 	}
@@ -221,7 +268,7 @@ func (a *Actuator) Update(ctx context.Context, cluster *clusterv1.Cluster, machi
 }
 
 // Exists tests for the existence of a machine and is invoked by the Machine Controller
-func (a *Actuator) Exists(ctx context.Context, cluster *clusterv1.Cluster, machine *machinev1.Machine) (bool, error) {
+func (a *Actuator) Exists(ctx context.Context, machine *machinev1beta1.Machine) (bool, error) {
 	log.Printf("Checking if machine %v exists.", machine.Name)
 	host, err := a.getHost(ctx, machine)
 	if err != nil {
@@ -240,13 +287,13 @@ func (a *Actuator) Exists(ctx context.Context, cluster *clusterv1.Cluster, machi
 // (https://github.com/kubernetes-sigs/cluster-api/issues/160).
 
 // GetIP returns IP address of the machine in the cluster.
-func (a *Actuator) GetIP(cluster *clusterv1.Cluster, machine *machinev1.Machine) (string, error) {
+func (a *Actuator) GetIP(machine *machinev1beta1.Machine) (string, error) {
 	log.Printf("Getting IP of machine %v .", machine.Name)
 	return "", fmt.Errorf("TODO: Not yet implemented")
 }
 
 // GetKubeConfig gets a kubeconfig from the running control plane.
-func (a *Actuator) GetKubeConfig(cluster *clusterv1.Cluster, controlPlaneMachine *machinev1.Machine) (string, error) {
+func (a *Actuator) GetKubeConfig(controlPlaneMachine *machinev1beta1.Machine) (string, error) {
 	log.Printf("Getting IP of machine %v .", controlPlaneMachine.Name)
 	return "", fmt.Errorf("TODO: Not yet implemented")
 }
@@ -254,7 +301,7 @@ func (a *Actuator) GetKubeConfig(cluster *clusterv1.Cluster, controlPlaneMachine
 // getHost gets the associated host by looking for an annotation on the machine
 // that contains a reference to the host. Returns nil if not found. Assumes the
 // host is in the same namespace as the machine.
-func (a *Actuator) getHost(ctx context.Context, machine *machinev1.Machine) (*bmh.BareMetalHost, error) {
+func (a *Actuator) getHost(ctx context.Context, machine *machinev1beta1.Machine) (*bmh.BareMetalHost, error) {
 	annotations := machine.ObjectMeta.GetAnnotations()
 	if annotations == nil {
 		return nil, nil
@@ -284,50 +331,60 @@ func (a *Actuator) getHost(ctx context.Context, machine *machinev1.Machine) (*bm
 	return &host, nil
 }
 
+// SelectorFromProviderSpec returns a selector that can be used to determine if
+// a BareMetalHost matches a Machine.
+func SelectorFromProviderSpec(providerspec *machinev1beta1.ProviderSpec) (labels.Selector, error) {
+	config, err := configFromProviderSpec(*providerspec)
+	if err != nil {
+		log.Printf("Error reading ProviderSpec: %s", err.Error())
+		return nil, err
+	}
+	selector := labels.NewSelector()
+	var reqs labels.Requirements
+	for labelKey, labelVal := range config.HostSelector.MatchLabels {
+		r, err := labels.NewRequirement(labelKey, selection.Equals, []string{labelVal})
+		if err != nil {
+			log.Printf("Failed to create MatchLabel requirement: %v", err)
+			return nil, err
+		}
+		reqs = append(reqs, *r)
+	}
+	for _, req := range config.HostSelector.MatchExpressions {
+		lowercaseOperator := selection.Operator(strings.ToLower(string(req.Operator)))
+		r, err := labels.NewRequirement(req.Key, lowercaseOperator, req.Values)
+		if err != nil {
+			log.Printf("Failed to create MatchExpression requirement: %v", err)
+			return nil, err
+		}
+		reqs = append(reqs, *r)
+	}
+	selector = selector.Add(reqs...)
+	return selector, nil
+}
+
 // chooseHost iterates through known hosts and returns one that can be
 // associated with the machine. It searches all hosts in case one already has an
 // association with this machine.
-func (a *Actuator) chooseHost(ctx context.Context, machine *machinev1.Machine,
-	config *bmv1alpha1.BareMetalMachineProviderSpec) (*bmh.BareMetalHost, error) {
-
+func (a *Actuator) chooseHost(ctx context.Context, machine *machinev1beta1.Machine) (*bmh.BareMetalHost, error) {
 	// get list of BMH
 	hosts := bmh.BareMetalHostList{}
 	opts := &client.ListOptions{
 		Namespace: machine.Namespace,
 	}
 
-	err := a.client.List(ctx, &hosts, client.UseListOptions(opts))
+	err := a.client.List(ctx, &hosts, opts)
 	if err != nil {
 		return nil, err
 	}
 
 	// Using the label selector on ListOptions above doesn't seem to work.
 	// I think it's because we have a local cache of all BareMetalHosts.
-	labelSelector := labels.NewSelector()
-	var reqs labels.Requirements
-	for labelKey, labelVal := range config.HostSelector.MatchLabels {
-		log.Printf("Adding requirement to match label: '%s' == '%s'", labelKey, labelVal)
-		r, err := labels.NewRequirement(labelKey, selection.Equals, []string{labelVal})
-		if err != nil {
-			log.Printf("Failed to create MatchLabel requirement, not choosing host: %v", err)
-			return nil, err
-		}
-		reqs = append(reqs, *r)
+	labelSelector, err := SelectorFromProviderSpec(&machine.Spec.ProviderSpec)
+	if err != nil {
+		return nil, err
 	}
-	for _, req := range config.HostSelector.MatchExpressions {
-		log.Printf("Adding requirement to match label: '%s' %s '%s'", req.Key, req.Operator, req.Values)
-		lowercaseOperator := selection.Operator(strings.ToLower(string(req.Operator)))
-		r, err := labels.NewRequirement(req.Key, lowercaseOperator, req.Values)
-		if err != nil {
-			log.Printf("Failed to create MatchExpression requirement, not choosing host: %v", err)
-			return nil, err
-		}
-		reqs = append(reqs, *r)
-	}
-	labelSelector = labelSelector.Add(reqs...)
 
 	availableHosts := []*bmh.BareMetalHost{}
-
 	for i, host := range hosts.Items {
 		if host.Available() {
 			if labelSelector.Matches(labels.Set(host.ObjectMeta.Labels)) {
@@ -355,7 +412,7 @@ func (a *Actuator) chooseHost(ctx context.Context, machine *machinev1.Machine,
 
 // consumerRefMatches returns a boolean based on whether the consumer
 // reference and machine metadata match
-func consumerRefMatches(consumer *corev1.ObjectReference, machine *machinev1.Machine) bool {
+func consumerRefMatches(consumer *corev1.ObjectReference, machine *machinev1beta1.Machine) bool {
 	if consumer.Name != machine.Name {
 		return false
 	}
@@ -374,7 +431,7 @@ func consumerRefMatches(consumer *corev1.ObjectReference, machine *machinev1.Mac
 // setHostSpec will ensure the host's Spec is set according to the machine's
 // details. It will then update the host via the kube API. If UserData does not
 // include a Namespace, it will default to the Machine's namespace.
-func (a *Actuator) setHostSpec(ctx context.Context, host *bmh.BareMetalHost, machine *machinev1.Machine,
+func (a *Actuator) setHostSpec(ctx context.Context, host *bmh.BareMetalHost, machine *machinev1beta1.Machine,
 	config *bmv1alpha1.BareMetalMachineProviderSpec) error {
 
 	// We only want to update the image setting if the host does not
@@ -407,7 +464,8 @@ func (a *Actuator) setHostSpec(ctx context.Context, host *bmh.BareMetalHost, mac
 
 // ensureAnnotation makes sure the machine has an annotation that references the
 // host and uses the API to update the machine if necessary.
-func (a *Actuator) ensureAnnotation(ctx context.Context, machine *machinev1.Machine, host *bmh.BareMetalHost) error {
+// it returns a boolean to indicate if the machine object was changed
+func (a *Actuator) ensureAnnotation(ctx context.Context, machine *machinev1beta1.Machine, host *bmh.BareMetalHost) (bool, error) {
 	annotations := machine.ObjectMeta.GetAnnotations()
 	if annotations == nil {
 		annotations = make(map[string]string)
@@ -415,26 +473,26 @@ func (a *Actuator) ensureAnnotation(ctx context.Context, machine *machinev1.Mach
 	hostKey, err := cache.MetaNamespaceKeyFunc(host)
 	if err != nil {
 		log.Printf("Error parsing annotation value \"%s\": %v", hostKey, err)
-		return err
+		return false, err
 	}
 	existing, ok := annotations[HostAnnotation]
 	if ok {
 		if existing == hostKey {
-			return nil
+			return false, nil
 		}
 		log.Printf("Warning: found stray annotation for host %s on machine %s. Overwriting.", existing, machine.Name)
 	}
 	annotations[HostAnnotation] = hostKey
 	machine.ObjectMeta.SetAnnotations(annotations)
-	return a.client.Update(ctx, machine)
+	return true, a.client.Update(ctx, machine)
 }
 
 // setError sets the ErrorMessage and ErrorReason fields on the machine and logs
 // the message. It assumes the reason is invalid configuration, since that is
 // currently the only relevant MachineStatusError choice.
-func (a *Actuator) setError(ctx context.Context, machine *machinev1.Machine, message string) error {
+func (a *Actuator) setError(ctx context.Context, machine *machinev1beta1.Machine, message string) error {
 	machine.Status.ErrorMessage = &message
-	reason := common.InvalidConfigurationMachineError
+	reason := machinev1beta1.InvalidConfigurationMachineError
 	machine.Status.ErrorReason = &reason
 	log.Printf("Machine %s: %s", machine.Name, message)
 	return a.client.Status().Update(ctx, machine)
@@ -443,7 +501,7 @@ func (a *Actuator) setError(ctx context.Context, machine *machinev1.Machine, mes
 // clearError removes the ErrorMessage from the machine's Status if set. Returns
 // nil if ErrorMessage was already nil. Returns a RequeueAfterError if the
 // machine was updated.
-func (a *Actuator) clearError(ctx context.Context, machine *machinev1.Machine) error {
+func (a *Actuator) clearError(ctx context.Context, machine *machinev1beta1.Machine) error {
 	if machine.Status.ErrorMessage != nil || machine.Status.ErrorReason != nil {
 		machine.Status.ErrorMessage = nil
 		machine.Status.ErrorReason = nil
@@ -452,14 +510,14 @@ func (a *Actuator) clearError(ctx context.Context, machine *machinev1.Machine) e
 			return err
 		}
 		log.Printf("Cleared error message from machine %s", machine.Name)
-		return &clustererror.RequeueAfterError{}
+		return &machineapierrors.RequeueAfterError{}
 	}
 	return nil
 }
 
 // configFromProviderSpec returns a BareMetalMachineProviderSpec by
 // deserializing the contents of a ProviderSpec
-func configFromProviderSpec(providerSpec machinev1.ProviderSpec) (*bmv1alpha1.BareMetalMachineProviderSpec, error) {
+func configFromProviderSpec(providerSpec machinev1beta1.ProviderSpec) (*bmv1alpha1.BareMetalMachineProviderSpec, error) {
 	if providerSpec.Value == nil {
 		return nil, fmt.Errorf("ProviderSpec missing")
 	}
@@ -474,7 +532,7 @@ func configFromProviderSpec(providerSpec machinev1.ProviderSpec) (*bmv1alpha1.Ba
 }
 
 // updateMachineStatus updates a machine object's status.
-func (a *Actuator) updateMachineStatus(ctx context.Context, machine *machinev1.Machine, host *bmh.BareMetalHost) error {
+func (a *Actuator) updateMachineStatus(ctx context.Context, machine *machinev1beta1.Machine, host *bmh.BareMetalHost) error {
 	addrs, err := a.nodeAddresses(host)
 	if err != nil {
 		return err
@@ -487,7 +545,7 @@ func (a *Actuator) updateMachineStatus(ctx context.Context, machine *machinev1.M
 	return nil
 }
 
-func (a *Actuator) applyMachineStatus(ctx context.Context, machine *machinev1.Machine, addrs []corev1.NodeAddress) error {
+func (a *Actuator) applyMachineStatus(ctx context.Context, machine *machinev1beta1.Machine, addrs []corev1.NodeAddress) error {
 	machineCopy := machine.DeepCopy()
 	machineCopy.Status.Addresses = addrs
 
@@ -534,3 +592,192 @@ func (a *Actuator) nodeAddresses(host *bmh.BareMetalHost) ([]corev1.NodeAddress,
 
 	return addrs, nil
 }
+
+//deleteRemediationAnnotations deletes poweredOffForRemediation and remediation strategy annotations
+func (a *Actuator) deleteRemediationAnnotations(ctx context.Context, machine *machinev1beta1.Machine) error {
+	if len(machine.Annotations) == 0 {
+		return nil
+	}
+
+	delete(machine.Annotations, poweredOffForRemediation)
+	delete(machine.Annotations, externalRemediationAnnotation)
+
+	if err := a.client.Update(ctx, machine); err != nil {
+		log.Printf("Failed to delete annotations of Machine: %s", machine.Name)
+		return err
+	}
+
+	return nil
+}
+
+//hasPowerOffRequestAnnotation checks if the requestPowerOffAnnotation exists on the baremetalhost
+func hasPowerOffRequestAnnotation(baremetalhost *bmh.BareMetalHost) (exists bool) {
+	if len(baremetalhost.Annotations) > 0 {
+		_, exists = baremetalhost.Annotations[requestPowerOffAnnotation]
+	}
+	return
+}
+
+//addPoweredOffForRemediationAnnotation adds a powered-off-for-remediation annotation to the machine
+func (a *Actuator) addPoweredOffForRemediationAnnotation(ctx context.Context, machine *machinev1beta1.Machine) error {
+	if machine.Annotations == nil {
+		machine.Annotations = make(map[string]string)
+	}
+
+	machine.Annotations[poweredOffForRemediation] = ""
+
+	err := a.client.Update(ctx, machine)
+	if err != nil {
+		log.Printf("Failed to add remediation in progess annotation to %s: %s", machine.Name, err.Error())
+	}
+
+	return err
+}
+
+//requestPowerOff adds requestPowerOffAnnotation on baremetalhost which signals BMO to power off the machine
+func (a *Actuator) requestPowerOff(ctx context.Context, baremetalhost *bmh.BareMetalHost) error {
+	if baremetalhost.Annotations == nil {
+		baremetalhost.Annotations = make(map[string]string)
+	}
+
+	if _, powerOffRequestExists := baremetalhost.Annotations[requestPowerOffAnnotation]; powerOffRequestExists {
+		return &machineapierrors.RequeueAfterError{RequeueAfter: time.Second * 5}
+	}
+
+	baremetalhost.Annotations[requestPowerOffAnnotation] = ""
+
+	err := a.client.Update(ctx, baremetalhost)
+	if err != nil {
+		log.Printf("failed to add power off request annotation to %s: %s", baremetalhost.Name, err.Error())
+	}
+
+	return err
+}
+
+//requestPowerOn removes requestPowerOffAnnotation from baremetalhost which signals BMO to power on the machine
+func (a *Actuator) requestPowerOn(ctx context.Context, baremetalhost *bmh.BareMetalHost) error {
+	if baremetalhost.Annotations == nil {
+		baremetalhost.Annotations = make(map[string]string)
+	}
+
+	if _, powerOffRequestExists := baremetalhost.Annotations[requestPowerOffAnnotation]; !powerOffRequestExists {
+		return &machineapierrors.RequeueAfterError{RequeueAfter: time.Second * 5}
+	}
+
+	delete(baremetalhost.Annotations, requestPowerOffAnnotation)
+
+	err := a.client.Update(ctx, baremetalhost)
+	if err != nil {
+		log.Printf("failed to power-off request annotation from %s: %s", baremetalhost.Name, err.Error())
+	}
+
+	return err
+}
+
+// deleteMachineNode deletes the node that mapped to specified machine
+func (a *Actuator) deleteNode(ctx context.Context, node *corev1.Node) error {
+	if !node.DeletionTimestamp.IsZero() {
+		return &machineapierrors.RequeueAfterError{RequeueAfter: time.Second * 2}
+	}
+
+	err := a.client.Delete(ctx, node)
+	if err != nil {
+		if errors.IsNotFound(err) {
+			return &machineapierrors.RequeueAfterError{}
+		}
+		log.Printf("Failed to delete node %s: %s", node.Name, err.Error())
+		return err
+	}
+	return &machineapierrors.RequeueAfterError{RequeueAfter: time.Second * 2}
+}
+
+// getNodeByMachine returns the node object referenced by machine
+func (a *Actuator) getNodeByMachine(ctx context.Context, machine *machinev1beta1.Machine) (*corev1.Node, error) {
+	if machine.Status.NodeRef == nil {
+		return nil, errors.NewNotFound(corev1.Resource("ObjectReference"), machine.Name)
+	}
+
+	node := &corev1.Node{}
+	key := client.ObjectKey{
+		Name:      machine.Status.NodeRef.Name,
+		Namespace: machine.Status.NodeRef.Namespace,
+	}
+
+	if err := a.client.Get(ctx, key, node); err != nil {
+		return nil, err
+	}
+	return node, nil
+}
+
+/*
+remediateIfNeeded will try to remediate unhealthy machines (annotated by MHC) by power-cycle the host
+The full remediation flow is:
+1) Power off the host
+2) Add poweredOffForRemediation annotation to the unhealthy Machine
+3) Delete the node
+4) Power on the host
+5) Wait for the node the come up (by waiting for the node to be registered in the cluster)
+6) Remove poweredOffForRemediation annotation and the MAO's machine unhealthy annotation
+*/
+func (a *Actuator) remediateIfNeeded(ctx context.Context, machine *machinev1beta1.Machine, baremetalhost *bmh.BareMetalHost) error {
+	if len(machine.Annotations) == 0 {
+		return nil
+	}
+
+	if _, needsRemediation := machine.Annotations[externalRemediationAnnotation]; !needsRemediation {
+		return nil
+	}
+
+	if _, poweredOffForRemediation := machine.Annotations[poweredOffForRemediation]; !poweredOffForRemediation {
+		if !hasPowerOffRequestAnnotation(baremetalhost) {
+			log.Printf("Found an unhealthy machine, requesting power off. Machine name: %s", machine.Name)
+			return a.requestPowerOff(ctx, baremetalhost)
+		}
+
+		//hold remediation until the power off request is fulfilled
+		if baremetalhost.Status.PoweredOn {
+			return nil
+		}
+
+		//we need this annotation to differentiate between unhealthy machine that
+		//needs remediation, and an unhealthy machine that just got remediated
+		return a.addPoweredOffForRemediationAnnotation(ctx, machine)
+	}
+
+	node, err := a.getNodeByMachine(ctx, machine)
+
+	if err != nil {
+		if !errors.IsNotFound(err) {
+			log.Printf("Failed to get Node from Machine %s: %s", machine.Name, err.Error())
+			return err
+		}
+	}
+
+	if !baremetalhost.Status.PoweredOn {
+		if node != nil {
+			log.Printf("Deleting Node %s associated with Machine %s", node.Name, machine.Name)
+			/*
+				Delete the node only after the host is powered off. Otherwise, if we would delete the node
+				when the host is powered on, the scheduler would assign the workload to other nodes, with the
+				possibility that two instances of the same application are running in parallel. This might result
+				in corruption or other issues for applications with singleton requirement. After the host is powered
+				off we know for sure that it is safe to re-assign that workload to other nodes.
+			*/
+			return a.deleteNode(ctx, node)
+		}
+
+		// node is deleted, we can power on the host
+		log.Printf("Requesting Host %s power on for Machine %s",
+			baremetalhost.Name, machine.Name)
+		return a.requestPowerOn(ctx, baremetalhost)
+	}
+
+	//node is still not running, so we requeue
+	if node == nil {
+		return &machineapierrors.RequeueAfterError{RequeueAfter: time.Second * 5}
+	}
+
+	// node is now available again
+	log.Printf("Node %s is available, remediation of Machine %s complete", node.Name, machine.Name)
+	return a.deleteRemediationAnnotations(ctx, machine)
+}
